@@ -18,7 +18,6 @@
 
 const Gtk = imports.gi.Gtk;
 const Clutter = imports.gi.Clutter;
-const Lang = imports.lang;
 const St = imports.gi.St;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
@@ -34,6 +33,7 @@ const GrabHelper = imports.ui.grabHelper;
 
 const ExtensionUtils = imports.misc.extensionUtils;
 const Me = ExtensionUtils.getCurrentExtension();
+const CreateFolderDialog = Me.imports.createFolderDialog;
 const Extension = Me.imports.extension;
 const FileItem = Me.imports.fileItem;
 const Prefs = Me.imports.prefs;
@@ -43,10 +43,7 @@ const Util = imports.misc.util;
 
 const Clipboard = St.Clipboard.get_default();
 const CLIPBOARD_TYPE = St.ClipboardType.CLIPBOARD;
-const Gettext = imports.gettext;
-
-Gettext.textdomain("desktop-icons");
-Gettext.bindtextdomain("desktop-icons", ExtensionUtils.getCurrentExtension().path + "/locale");
+const Gettext = imports.gettext.domain('desktop-icons');
 
 const _ = Gettext.gettext;
 
@@ -58,6 +55,11 @@ var UndoStatus = {
     REDO: 2,
 };
 
+var StoredCoordinates = {
+    PRESERVE: 0,
+    OVERWRITE:1,
+};
+
 class Placeholder extends St.Bin {
     constructor() {
         super();
@@ -78,7 +80,14 @@ var DesktopGrid = class {
             row_homogeneous: true
         });
 
-        this.actor = new St.Widget({ layout_manager: new Clutter.FixedLayout() });
+        this._actorLayout = new Clutter.BinLayout({
+            x_align: Clutter.BinAlignment.FIXED,
+            y_align: Clutter.BinAlignment.FIXED
+        });
+
+        this.actor = new St.Widget({
+            layout_manager: this._actorLayout
+        });
         this.actor._delegate = this;
 
         this._grid = new St.Widget({
@@ -90,12 +99,12 @@ var DesktopGrid = class {
             can_focus: true,
             opacity: 255
         });
-        this.actor.add_actor(this._grid);
+        this.actor.add_child(this._grid);
 
         this._renamePopup = new RenamePopup(this);
-        this.actor.add_actor(this._renamePopup.actor);
+        this.actor.add_child(this._renamePopup.actor);
 
-        this._bgManager._container.add_actor(this.actor);
+        this._bgManager._container.add_child(this.actor);
 
         this.actor.connect('destroy', () => this._onDestroy());
 
@@ -112,13 +121,10 @@ var DesktopGrid = class {
             () => this._backgroundDestroyed());
 
         this._grid.connect('button-press-event', (actor, event) => this._onPressButton(actor, event));
-        this._rubberBand = new St.Widget({ style_class: 'rubber-band' });
-        this._rubberBand.hide();
-        Main.layoutManager.uiGroup.add_actor(this._rubberBand);
 
         this._grid.connect('key-press-event', this._onKeyPress.bind(this));
 
-        this.reset();
+        this._grid.connect('allocation-changed', () => Extension.desktopManager.scheduleReLayoutChildren());
     }
 
     _onKeyPress(actor, event) {
@@ -155,6 +161,10 @@ var DesktopGrid = class {
         else if (symbol == Clutter.Delete) {
             Extension.desktopManager.doTrash();
             return Clutter.EVENT_STOP;
+        } else if (symbol == Clutter.F2) {
+            // Support renaming other grids file items.
+            Extension.desktopManager.doRename();
+            return Clutter.EVENT_STOP;
         }
 
         return Clutter.EVENT_PROPAGATE;
@@ -178,17 +188,23 @@ var DesktopGrid = class {
             this._bgManager.backgroundActor.disconnect(this._bgDestroyedId);
         this._bgDestroyedId = 0;
         this._bgManager = null;
-        this._rubberBand.destroy();
     }
 
-    _omNewFolderClicked() {
-        let dir = DesktopIconsUtil.getDesktopDir().get_child(_('New Folder'));
-        DBusUtils.NautilusFileOperationsProxy.CreateFolderRemote(dir.get_uri(),
-            (result, error) => {
-                if (error)
-                    throw new Error('Error creating new folder: ' + error.message);
-            }
-        );
+    _onNewFolderClicked() {
+
+        let dialog = new CreateFolderDialog.CreateFolderDialog();
+
+        dialog.connect('response', (dialog, name) => {
+            let dir = DesktopIconsUtil.getDesktopDir().get_child(name);
+            DBusUtils.NautilusFileOperationsProxy.CreateFolderRemote(dir.get_uri(),
+                (result, error) => {
+                    if (error)
+                        throw new Error('Error creating new folder: ' + error.message);
+                }
+            );
+        });
+
+        dialog.open();
     }
 
     _parseClipboardText(text) {
@@ -271,9 +287,9 @@ var DesktopGrid = class {
     _onOpenDesktopInFilesClicked() {
         Gio.AppInfo.launch_default_for_uri_async(DesktopIconsUtil.getDesktopDir().get_uri(),
             null, null,
-            (source, res) => {
+            (source, result) => {
                 try {
-                    Gio.AppInfo.launch_default_for_uri_finish(res);
+                    Gio.AppInfo.launch_default_for_uri_finish(result);
                 } catch (e) {
                    log('Error opening Desktop in Files: ' + e.message);
                 }
@@ -282,8 +298,8 @@ var DesktopGrid = class {
     }
 
     _onOpenTerminalClicked() {
-        let desktopUri = DesktopIconsUtil.getDesktopDir().get_uri();
-        let command = DesktopIconsUtil.getTerminalCommand(desktopUri);
+        let desktopPath = DesktopIconsUtil.getDesktopDir().get_path();
+        let command = DesktopIconsUtil.getTerminalCommand(desktopPath);
 
         Util.spawnCommandLine(command);
     }
@@ -301,22 +317,23 @@ var DesktopGrid = class {
     _createDesktopBackgroundMenu() {
         let menu = new PopupMenu.PopupMenu(Main.layoutManager.dummyCursor,
                                            0, St.Side.TOP);
-        menu.addAction(_('New Folder'), () => this._omNewFolderClicked());
+        menu.addAction(_("New Folder"), () => this._onNewFolderClicked());
         menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
-        this._pasteMenuItem = menu.addAction(_('Paste'), () => this._onPasteClicked());
-        this._undoMenuItem = menu.addAction(_('Undo'), () => this._onUndoClicked());
-        this._redoMenuItem = menu.addAction(_('Redo'), () => this._onRedoClicked());
+        this._pasteMenuItem = menu.addAction(_("Paste"), () => this._onPasteClicked());
+        this._undoMenuItem = menu.addAction(_("Undo"), () => this._onUndoClicked());
+        this._redoMenuItem = menu.addAction(_("Redo"), () => this._onRedoClicked());
         menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
-        menu.addAction(_('Open Desktop in Files'), () => this._onOpenDesktopInFilesClicked());
-        menu.addAction(_('Open Terminal'), () => this._onOpenTerminalClicked());
+        menu.addAction(_("Show Desktop in Files"), () => this._onOpenDesktopInFilesClicked());
+        menu.addAction(_("Open in Terminal"), () => this._onOpenTerminalClicked());
         menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
-        menu.addSettingsAction(_('Change Background…'), 'gnome-background-panel.desktop');
-        menu.addSettingsAction(_('Display Settings'), 'gnome-display-panel.desktop');
-        menu.addSettingsAction(_('Settings'), 'gnome-control-center.desktop');
+        menu.addSettingsAction(_("Change Background…"), 'gnome-background-panel.desktop');
+        menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
+        menu.addSettingsAction(_("Display Settings"), 'gnome-display-panel.desktop');
+        menu.addSettingsAction(_("Settings"), 'gnome-control-center.desktop');
 
         menu.actor.add_style_class_name('background-menu');
 
-        Main.layoutManager.uiGroup.add_actor(menu.actor);
+        Main.layoutManager.uiGroup.add_child(menu.actor);
         menu.actor.hide();
 
         menu._propertiesChangedId = DBusUtils.NautilusFileOperationsProxy.connect('g-properties-changed',
@@ -350,52 +367,41 @@ var DesktopGrid = class {
         this.actor._desktopBackgroundManager.ignoreRelease();
     }
 
-    dropItems(fileItems) {
-        let reserved = {};
-        for (let fileItem of fileItems) {
-            let [dropX, dropY] = (fileItem.savedCoordinates == null) ? [0, 0] : fileItem.savedCoordinates;
-            let [column, row] = this._getEmptyPlaceClosestTo(dropX, dropY, reserved);
-            let placeholder = this.layout.get_child_at(column, row);
-            let hashedPosition = `${column},${row}`;
-            if (hashedPosition in reserved)
-                continue;
-
-            reserved[`${column},${row}`] = fileItem;
-            placeholder.child = fileItem.actor;
-            this._addFileItemTo(fileItem, column, row);
-        }
-    }
-
-    _addFileItemTo(fileItem, column, row) {
+    _addFileItemTo(fileItem, column, row, overwriteCoordinates) {
         let placeholder = this.layout.get_child_at(column, row);
         placeholder.child = fileItem.actor;
         this._fileItems.push(fileItem);
         let selectedId = fileItem.connect('selected', this._onFileItemSelected.bind(this));
-        let renameId = fileItem.connect('rename-clicked', this._onFileItemRenameClicked.bind(this));
+        let renameId = fileItem.connect('rename-clicked', this.doRename.bind(this));
         this._fileItemHandlers.set(fileItem, [selectedId, renameId]);
+
         /* If this file is new in the Desktop and hasn't yet
          * fixed coordinates, store the new possition to ensure
-         * that the next time it will be shown in the same possition
+         * that the next time it will be shown in the same possition.
+         * Also store the new possition if it has been moved by the user,
+         * and not triggered by a screen change.
          */
-        if (fileItem.savedCoordinates == null) {
+        if ((fileItem.savedCoordinates == null) || (overwriteCoordinates == StoredCoordinates.OVERWRITE)) {
             let [fileX, fileY] = placeholder.get_transformed_position();
             fileItem.savedCoordinates = [Math.round(fileX), Math.round(fileY)];
         }
     }
 
-    addFileItemCloseTo(fileItem, x, y) {
-        let [column, row] = this._getEmptyPlaceClosestTo(x, y, null);
-        this._addFileItemTo(fileItem, column, row);
+    addFileItemCloseTo(fileItem, x, y, overwriteCoordinates) {
+        let [column, row] = this._getEmptyPlaceClosestTo(x, y);
+        this._addFileItemTo(fileItem, column, row, overwriteCoordinates);
     }
 
-    _getEmptyPlaceClosestTo(x, y, reserved) {
+    _getEmptyPlaceClosestTo(x, y) {
         let maxColumns = this._getMaxColumns();
         let maxRows = this._getMaxRows();
 
-        let workarea = Main.layoutManager.getWorkAreaForMonitor(this._monitorConstraint.index);
-        let [actorX, actorY] = this.actor.get_transformed_position();
-        let placeX = Math.round((x - actorX) * maxColumns / workarea.width);
-        let placeY = Math.round((y - actorY) * maxRows / workarea.height);
+        let [actorX, actorY] = this._grid.get_transformed_position();
+        let actorWidth = this._grid.allocation.x2 - this._grid.allocation.x1;
+        let actorHeight = this._grid.allocation.y2 - this._grid.allocation.y1;
+        let placeX = Math.round((x - actorX) * maxColumns / actorWidth);
+        let placeY = Math.round((y - actorY) * maxRows / actorHeight);
+
         placeX = DesktopIconsUtil.clamp(placeX, 0, maxColumns - 1);
         placeY = DesktopIconsUtil.clamp(placeY, 0, maxRows - 1);
         if (this.layout.get_child_at(placeX, placeY).child == null)
@@ -410,9 +416,6 @@ var DesktopGrid = class {
                 if (placeholder.child != null)
                     continue;
 
-                if (reserved && `${column},${row}` in reserved)
-                    continue;
-
                 let [proposedX, proposedY] = placeholder.get_transformed_position();
                 let distance = DesktopIconsUtil.distanceBetweenPoints(proposedX, proposedY, x, y);
                 if (distance < minDistance) {
@@ -455,7 +458,8 @@ var DesktopGrid = class {
     }
 
     reset() {
-        for (let fileItem of this._fileItems)
+        let tmpFileItemsCopy = this._fileItems.slice();
+        for (let fileItem of tmpFileItemsCopy)
             this.removeFileItem(fileItem);
         this._grid.remove_all_children();
 
@@ -474,12 +478,15 @@ var DesktopGrid = class {
     _onPressButton(actor, event) {
         let button = event.get_button();
         let [x, y] = event.get_coords();
+
+        this._grid.grab_key_focus();
+
         if (button == 1) {
             let shiftPressed = !!(event.get_state() & Clutter.ModifierType.SHIFT_MASK);
             let controlPressed = !!(event.get_state() & Clutter.ModifierType.CONTROL_MASK);
             if (!shiftPressed && !controlPressed)
                 Extension.desktopManager.clearSelection();
-            let [gridX, gridY] = this.actor.get_transformed_position();
+            let [gridX, gridY] = this._grid.get_transformed_position();
             Extension.desktopManager.startRubberBand(x, y, gridX, gridY);
             return Clutter.EVENT_STOP;
         }
@@ -506,13 +513,13 @@ var DesktopGrid = class {
     }
 
     _getMaxColumns() {
-        let workarea = Main.layoutManager.getWorkAreaForMonitor(this._monitorConstraint.index);
-        return Math.floor(workarea.width / Prefs.get_desired_width(St.ThemeContext.get_for_stage(global.stage).scale_factor));
+        let gridWidth = this._grid.allocation.x2 - this._grid.allocation.x1;
+        return Math.floor(gridWidth / Prefs.get_desired_width(St.ThemeContext.get_for_stage(global.stage).scale_factor));
     }
 
     _getMaxRows() {
-        let workarea = Main.layoutManager.getWorkAreaForMonitor(this._monitorConstraint.index);
-        return Math.floor(workarea.height / Prefs.get_desired_height(St.ThemeContext.get_for_stage(global.stage).scale_factor));
+        let gridHeight = this._grid.allocation.y2 - this._grid.allocation.y1;
+        return Math.floor(gridHeight / Prefs.get_desired_height(St.ThemeContext.get_for_stage(global.stage).scale_factor));
     }
 
     acceptDrop(source, actor, x, y, time) {
@@ -555,7 +562,7 @@ var DesktopGrid = class {
         this._grid.grab_key_focus();
     }
 
-    _onFileItemRenameClicked(fileItem) {
+    doRename(fileItem) {
         this._renamePopup.onFileItemRenameClicked(fileItem);
     }
 };
@@ -570,7 +577,7 @@ var RenamePopup = class {
                                            can_focus: true,
                                            x_expand: true });
         this._renameEntry.clutter_text.connect('activate', this._onRenameAccepted.bind(this));
-        this._renameOkButton= new St.Button({ label: _("Ok"),
+        this._renameOkButton= new St.Button({ label: _("OK"),
                                               style_class: 'app-view-control button',
                                               button_mask: St.ButtonMask.ONE | St.ButtonMask.THREE,
                                               reactive: true,
@@ -587,18 +594,18 @@ var RenamePopup = class {
         let renameButtonsBoxLayout = new Clutter.BoxLayout({ homogeneous: true });
         let renameButtonsBox = new St.Widget({ layout_manager: renameButtonsBoxLayout,
                                                x_expand: true });
-        renameButtonsBox.add_actor(this._renameCancelButton);
-        renameButtonsBox.add_actor(this._renameOkButton);
+        renameButtonsBox.add_child(this._renameCancelButton);
+        renameButtonsBox.add_child(this._renameOkButton);
 
         let renameContentLayout = new Clutter.BoxLayout({ spacing: 6,
                                                           orientation: Clutter.Orientation.VERTICAL });
         let renameContent = new St.Widget({ style_class: 'rename-popup',
                                             layout_manager: renameContentLayout,
                                             x_expand: true });
-        renameContent.add_actor(this._renameEntry);
-        renameContent.add_actor(renameButtonsBox);
+        renameContent.add_child(this._renameEntry);
+        renameContent.add_child(renameButtonsBox);
 
-        this._boxPointer = new BoxPointer.BoxPointer(St.Side.TOP, { can_focus: false, x_expand: true });
+        this._boxPointer = new BoxPointer.BoxPointer(St.Side.TOP, { can_focus: false, x_expand: false });
         this.actor = this._boxPointer.actor;
         this.actor.style_class = 'popup-menu-boxpointer';
         this.actor.add_style_class_name('popup-menu');
@@ -648,9 +655,9 @@ var RenamePopup = class {
 
         this._popup();
         this._renameEntry.grab_key_focus();
-        this._renameEntry.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false)
-        let allChars = fileItem.displayName.length;
-        this._renameEntry.clutter_text.set_selection(0, allChars);
+        this._renameEntry.navigate_focus(null, Gtk.DirectionType.TAB_FORWARD, false);
+        let extensionOffset = DesktopIconsUtil.getFileExtensionOffset(fileItem.displayName, fileItem.isDirectory);
+        this._renameEntry.clutter_text.set_selection(0, extensionOffset);
     }
 
     _onRenameAccepted() {
