@@ -16,16 +16,15 @@
  * along with this program.  If not, see <http://www.gnu.org/licenses/>.
  */
 
+const Gtk = imports.gi.Gtk;
 const Clutter = imports.gi.Clutter;
+const GObject = imports.gi.GObject;
 const Gio = imports.gi.Gio;
 const GLib = imports.gi.GLib;
-const Lang = imports.lang;
 const St = imports.gi.St;
 const Mainloop = imports.mainloop;
 const Meta = imports.gi.Meta;
 
-const Signals = imports.signals;
-
 const Animation = imports.ui.animation;
 const Background = imports.ui.background;
 const DND = imports.ui.dnd;
@@ -44,6 +43,8 @@ const DesktopIconsUtil = Me.imports.desktopIconsUtil;
 const Clipboard = St.Clipboard.get_default();
 const CLIPBOARD_TYPE = St.ClipboardType.CLIPBOARD;
 
+var S_IWOTH = 0x00002;
+
 function getDpy() {
     return global.screen || global.display;
 }
@@ -53,27 +54,46 @@ function findMonitorIndexForPos(x, y) {
 }
 
 
-var DesktopManager = class {
-    constructor() {
+var DesktopManager = GObject.registerClass({
+    Properties: {
+        'writable-by-others': GObject.ParamSpec.boolean(
+            'writable-by-others',
+            'WritableByOthers',
+            'Whether the desktop\'s directory can be written by others (o+w unix permission)',
+            GObject.ParamFlags.READABLE,
+            false
+        )
+    }
+}, class DesktopManager extends GObject.Object {
+    _init(params) {
+        super._init(params);
+
         this._layoutChildrenId = 0;
-        this._scheduleDesktopsRefreshId = 0;
+        this._deleteChildrenId = 0;
         this._monitorDesktopDir = null;
         this._desktopMonitorCancellable = null;
         this._desktopGrids = {};
         this._fileItemHandlers = new Map();
-        this._fileItems = [];
+        this._fileItems = new Map();
         this._dragCancelled = false;
+        this._queryFileInfoCancellable = null;
+        this._unixMode = null;
+        this._writableByOthers = null;
 
         this._monitorsChangedId = Main.layoutManager.connect('monitors-changed', () => this._recreateDesktopIcons());
         this._rubberBand = new St.Widget({ style_class: 'rubber-band' });
         this._rubberBand.hide();
-        Main.layoutManager.uiGroup.add_actor(this._rubberBand);
+        Main.layoutManager.uiGroup.add_child(this._rubberBand);
         this._grabHelper = new GrabHelper.GrabHelper(global.stage);
 
         this._addDesktopIcons();
         this._monitorDesktopFolder();
 
-        Prefs.settings.connect("changed", () => this._recreateDesktopIcons());
+        this.settingsId = Prefs.settings.connect('changed', () => this._recreateDesktopIcons());
+        this.gtkSettingsId = Prefs.gtkSettings.connect('changed', (obj, key) => {
+            if (key == 'show-hidden')
+                this._recreateDesktopIcons();
+        });
 
         this._selection = new Set();
         this._inDrag = false;
@@ -84,6 +104,7 @@ var DesktopManager = class {
     startRubberBand(x, y) {
         this._rubberBandInitialX = x;
         this._rubberBandInitialY = y;
+        this._initRubberBandColor();
         this._updateRubberBand(x, y);
         this._rubberBand.show();
         this._grabHelper.grab({ actor: global.stage });
@@ -92,6 +113,20 @@ var DesktopManager = class {
             this.endRubberBand();
         });
         this._rubberBandId = global.stage.connect('motion-event', (actor, event) => {
+            /* In some cases, when the user starts a rubberband selection and ends it
+             * (by releasing the left button) over a window instead of doing it over
+             * the desktop, the stage doesn't receive the "button-release" event.
+             * This happens currently with, at least, Dash to Dock extension, but
+             * it probably also happens with other applications or extensions.
+             * To fix this, we also end the rubberband selection if we detect mouse
+             * motion in the stage without the left button pressed during a
+             * rubberband selection.
+             *  */
+            let button = event.get_state();
+            if (!(button & Clutter.ModifierType.BUTTON1_MASK)) {
+                this.endRubberBand();
+                return;
+            }
             [x, y] = event.get_coords();
             this._updateRubberBand(x, y);
             let x0, y0, x1, y1;
@@ -109,7 +144,7 @@ var DesktopManager = class {
                 y1 = this._rubberBandInitialY;
                 y0 = y;
             }
-            for(let fileItem of this._fileItems) {
+            for (let [fileUri, fileItem] of this._fileItems) {
                 fileItem.emit('selected', true,
                               fileItem.intersectsWith(x0, y0, x1 - x0, y1 - y0));
             }
@@ -146,9 +181,9 @@ var DesktopManager = class {
     _addDesktopIcons() {
         forEachBackgroundManager(bgManager => {
             let newGrid = new DesktopGrid.DesktopGrid(bgManager);
-            newGrid.actor.connect("destroy", (actor) => {
+            newGrid.actor.connect('destroy', (actor) => {
                 // if a grid loses its actor, remove it from the grid list
-                for(let grid in this._desktopGrids)
+                for (let grid in this._desktopGrids)
                     if (this._desktopGrids[grid].actor == actor) {
                         delete this._desktopGrids[grid];
                         break;
@@ -165,32 +200,56 @@ var DesktopManager = class {
         this._desktopGrids = {};
     }
 
+    /** 
+     * Initialize rubberband color from the GTK rubberband class
+     * */
+    _initRubberBandColor() {
+        let rgba = DesktopIconsUtil.getGtkClassBackgroundColor('rubberband', Gtk.StateFlags.NORMAL);
+        let background_color =
+            'rgba(' + rgba.red * 255 + ', ' + rgba.green * 255 + ', ' + rgba.blue * 255 + ', 0.4)';
+        this._rubberBand.set_style('background-color: ' + background_color);
+    }
+
     async _scanFiles() {
         for (let [fileItem, id] of this._fileItemHandlers)
             fileItem.disconnect(id);
         this._fileItemHandlers = new Map();
-        this._fileItems = [];
+
+        if (!this._unixMode) {
+            let desktopDir = DesktopIconsUtil.getDesktopDir();
+            let fileInfo = desktopDir.query_info(Gio.FILE_ATTRIBUTE_UNIX_MODE,
+                                                 Gio.FileQueryInfoFlags.NONE,
+                                                 null);
+            this._unixMode = fileInfo.get_attribute_uint32(Gio.FILE_ATTRIBUTE_UNIX_MODE);
+            this._setWritableByOthers((this._unixMode & S_IWOTH) != 0);
+        }
 
         try {
+            let tmpFileItems = new Map();
             for (let [file, info, extra] of await this._enumerateDesktop()) {
                 let fileItem = new FileItem.FileItem(file, info, extra);
-                this._fileItems.push(fileItem);
+                tmpFileItems.set(fileItem.file.get_uri(), fileItem);
                 let id = fileItem.connect('selected',
                                           this._onFileItemSelected.bind(this));
 
                 this._fileItemHandlers.set(fileItem, id);
             }
+            this._fileItems = tmpFileItems;
         } catch (e) {
             if (!e.matches(Gio.IOErrorEnum, Gio.IOErrorEnum.CANCELLED))
                 log(`Error loading desktop files ${e.message}`);
             return;
         }
 
-        Object.values(this._desktopGrids).forEach(grid => {
-            grid.actor.connect('allocation-changed', () => this._scheduleLayoutChildren());
-        });
+        this.scheduleReLayoutChildren();
+    }
 
-        this._scheduleReLayoutChildren();
+    getDesktopFileNames () {
+        let fileList = [];
+        for (let [uri, item] of this._fileItems) {
+            fileList.push(item.fileName);
+        }
+        return fileList;
     }
 
     _enumerateDesktop() {
@@ -205,16 +264,16 @@ var DesktopManager = class {
                 Gio.FileQueryInfoFlags.NONE,
                 GLib.PRIORITY_DEFAULT,
                 this._desktopEnumerateCancellable,
-                (o, res) => {
+                (source, result) => {
                     try {
-                        let fileEnum = desktopDir.enumerate_children_finish(res);
+                        let fileEnum = source.enumerate_children_finish(result);
                         let resultGenerator = function *() {
                             let info;
-                            while ((info = fileEnum.next_file(null)))
-                                yield [fileEnum.get_child(info), info, Prefs.FILE_TYPE.NONE];
                             for (let [newFolder, extras] of DesktopIconsUtil.getExtraFolders()) {
                                 yield [newFolder, newFolder.query_info(DesktopIconsUtil.DEFAULT_ATTRIBUTES, Gio.FileQueryInfoFlags.NONE, this._desktopEnumerateCancellable), extras];
                             }
+                            while ((info = fileEnum.next_file(null)))
+                                yield [fileEnum.get_child(info), info, Prefs.FileType.NONE];
                         }.bind(this);
                         resolve(resultGenerator());
                     } catch (e) {
@@ -236,30 +295,86 @@ var DesktopManager = class {
         this._monitorDesktopDir.connect('changed', (obj, file, otherFile, eventType) => this._updateDesktopIfChanged(file, otherFile, eventType));
     }
 
+    checkIfSpecialFilesAreSelected() {
+        for (let fileItem of this._selection) {
+            if (fileItem.isSpecial)
+                return true;
+        }
+        return false;
+    }
+
+    getNumberOfSelectedItems() {
+        return this._selection.size;
+    }
+
+    get writableByOthers() {
+        return this._writableByOthers;
+    }
+
+    _setWritableByOthers(value) {
+        if (value == this._writableByOthers)
+            return;
+
+        this._writableByOthers = value
+        this.notify('writable-by-others');
+    }
+
     _updateDesktopIfChanged (file, otherFile, eventType) {
+        let {
+            DELETED, MOVED_IN, MOVED_OUT, CREATED, RENAMED, CHANGES_DONE_HINT, ATTRIBUTE_CHANGED
+        } = Gio.FileMonitorEvent;
+
+        let fileUri = file.get_uri();
+        let fileItem = null;
+        if (this._fileItems.has(fileUri))
+            fileItem = this._fileItems.get(fileUri);
+        switch(eventType) {
+            case RENAMED:
+                this._fileItems.delete(fileUri);
+                this._fileItems.set(otherFile.get_uri(), fileItem);
+                fileItem.onFileRenamed(otherFile);
+                return;
+            case CHANGES_DONE_HINT:
+            case ATTRIBUTE_CHANGED:
+                /* a file changed, rather than the desktop itself */
+                let desktopDir = DesktopIconsUtil.getDesktopDir();
+                if (file.get_uri() != desktopDir.get_uri())
+                    return;
+
+                if (this._queryFileInfoCancellable)
+                    this._queryFileInfoCancellable.cancel();
+
+                file.query_info_async(Gio.FILE_ATTRIBUTE_UNIX_MODE,
+                                      Gio.FileQueryInfoFlags.NONE,
+                                      GLib.PRIORITY_DEFAULT,
+                                      this._queryFileInfoCancellable,
+                    (source, result) => {
+                        try {
+                            let info = source.query_info_finish(result);
+                            this._queryFileInfoCancellable = null;
+
+                            this._unixMode = info.get_attribute_uint32(Gio.FILE_ATTRIBUTE_UNIX_MODE);
+                            this._setWritableByOthers((this._unixMode & S_IWOTH) != 0);
+
+                            if (this._writableByOthers)
+                                log(`desktop-icons: Desktop is writable by others - will not allow launching any desktop files`);
+                        } catch(error) {
+                            if (!error.matches(Gio.IOErrorEnum, Gio.IOErrorEnum.CANCELLED))
+                                global.log('Error getting desktop unix mode: ' + error);
+                        }
+                    });
+
+                return;
+        }
+
         // Only get a subset of events we are interested in.
         // Note that CREATED will emit a CHANGES_DONE_HINT
-        let {
-            CHANGES_DONE_HINT, DELETED, RENAMED, MOVED_IN, MOVED_OUT, CREATED
-        } = Gio.FileMonitorEvent;
-        if (![CHANGES_DONE_HINT, DELETED, RENAMED,
-            MOVED_IN, MOVED_OUT, CREATED].includes(eventType))
+        if (![DELETED, MOVED_IN, MOVED_OUT, CREATED].includes(eventType))
             return;
 
         this._recreateDesktopIcons();
     }
 
-    _getContainerWithChild(child) {
-        let monitorIndex = Main.layoutManager.findIndexForActor(child);
-        let desktopGrid = this._desktopGrids[monitorIndex];
-        let children = desktopGrid.actor.get_children();
-
-        if (children.some(x => x.child == child))
-            return desktopGrid;
-        else
-            throw new Error("Missmatch between expected items in a desktop grid not found");
-    }
-
     _setupDnD() {
         this._draggableContainer = new St.Widget({
             visible: true,
@@ -300,7 +415,7 @@ var DesktopManager = class {
             });
             clone.x = fileItem.actor.get_transformed_position()[0];
             clone.y = fileItem.actor.get_transformed_position()[1];
-            this._draggableContainer.add_actor(clone);
+            this._draggableContainer.add_child(clone);
         }
 
         Main.layoutManager.uiGroup.add_child(this._draggableContainer);
@@ -396,76 +511,77 @@ var DesktopManager = class {
          * them.
          */
         for (let item of this._selection) {
-            let [itemX, itemY] = (item.savedCoordinates == null) ? [0, 0] : item.savedCoordinates;
+            let [itemX, itemY] = item.actor.get_transformed_position();
             let monitorIndex = findMonitorIndexForPos(itemX, itemY);
-            savedCoordinates.set(item, item.actor.get_transformed_position());
+            savedCoordinates.set(item, [itemX, itemY]);
             this._desktopGrids[monitorIndex].removeFileItem(item);
-            /* Set the new ideal position where the item drop should happen */
-            let newfileX = Math.round(xDiff + itemX);
-            let newfileY = Math.round(yDiff + itemY);
-            item.savedCoordinates = [newfileX, newfileY];
         }
 
-        for (let key in this._desktopGrids) {
-            /* Create list of items to drop per desktop. We want to drop them at
-             * once so calculations for the dropping per grid are smoother
-             */
-            let itemsForDropDesktop = [...this._selection].filter(
-                (x) => {
-                    let [itemX, itemY] = (x.savedCoordinates == null) ? [0, 0] : x.savedCoordinates;
-                    let monitorIndex = findMonitorIndexForPos(itemX, itemY);
-                    return key == monitorIndex;
-                }
-            );
-
-            this._desktopGrids[key].dropItems(itemsForDropDesktop);
+        for (let item of this._selection) {
+            let [itemX, itemY] = savedCoordinates.get(item);
+            /* Set the new ideal position where the item drop should happen */
+            let newFileX = Math.round(xDiff + itemX);
+            let newFileY = Math.round(yDiff + itemY);
+            let monitorIndex = findMonitorIndexForPos(newFileX, newFileY);
+            this._desktopGrids[monitorIndex].addFileItemCloseTo(item, newFileX, newFileY, DesktopGrid.StoredCoordinates.OVERWRITE);
         }
 
         return true;
     }
 
-    selectionDropOnFileItem (fileItem) {
-        if (!fileItem.isDirectory)
+    selectionDropOnFileItem (fileItemDestination) {
+        if (!fileItemDestination.isDirectory)
             return false;
 
         let droppedUris = [];
         for (let fileItem of this._selection) {
             if (fileItem.isSpecial)
                 return false;
+            if (fileItemDestination.file.get_uri() == fileItem.file.get_uri())
+                return false;
             droppedUris.push(fileItem.file.get_uri());
         }
 
+        if (droppedUris.length == 0)
+            return true;
+
         DBusUtils.NautilusFileOperationsProxy.MoveURIsRemote(droppedUris,
-                                                             fileItem.file.get_uri(),
+                                                             fileItemDestination.file.get_uri(),
             (result, error) => {
                 if (error)
                     throw new Error('Error moving files: ' + error.message);
             }
         );
-
         for (let fileItem of this._selection) {
             fileItem.state = FileItem.State.GONE;
         }
 
-        this._scheduleReLayoutChildren();
+        this._recreateDesktopIcons();
 
         return true;
     }
 
-    _scheduleLayoutChildren() {
-        if (this._layoutChildrenId != 0)
-            return;
-
-        this._layoutChildrenId = GLib.idle_add(GLib.PRIORITY_LOW, () => this._layoutChildren());
-    }
-
-    _scheduleReLayoutChildren() {
-        if (this._layoutChildrenId != 0)
-            return;
+    _resetGridsAndScheduleLayout() {
+        this._deleteChildrenId = 0;
 
         Object.values(this._desktopGrids).forEach((grid) => grid.reset());
 
         this._layoutChildrenId = GLib.idle_add(GLib.PRIORITY_LOW, () => this._layoutChildren());
+
+        return GLib.SOURCE_REMOVE;
+    }
+
+    scheduleReLayoutChildren() {
+        if (this._deleteChildrenId != 0)
+            return;
+
+        if (this._layoutChildrenId != 0) {
+            GLib.source_remove(this._layoutChildrenId);
+            this._layoutChildrenId = 0;
+        }
+
+
+        this._deleteChildrenId = GLib.idle_add(GLib.PRIORITY_LOW, () => this._resetGridsAndScheduleLayout());
     }
 
     _addFileItemCloseTo(item) {
@@ -473,31 +589,36 @@ var DesktopManager = class {
         let monitorIndex = findMonitorIndexForPos(x, y);
         let desktopGrid = this._desktopGrids[monitorIndex];
         try {
-            desktopGrid.addFileItemCloseTo(item, x, y);
+            desktopGrid.addFileItemCloseTo(item, x, y, DesktopGrid.StoredCoordinates.PRESERVE);
         } catch (e) {
             log(`Error adding children to desktop: ${e.message}`);
         }
     }
 
     _layoutChildren() {
+        let showHidden = Prefs.gtkSettings.get_boolean('show-hidden');
         /*
          * Paint the icons in two passes:
          * * first pass paints those that have their coordinates defined in the metadata
          * * second pass paints those new files that still don't have their definitive coordinates
          */
-        for (let fileItem of this._fileItems) {
+        for (let [fileUri, fileItem] of this._fileItems) {
             if (fileItem.savedCoordinates == null)
                 continue;
             if (fileItem.state != FileItem.State.NORMAL)
                 continue;
+            if (!showHidden && fileItem.isHidden)
+                continue;
             this._addFileItemCloseTo(fileItem);
         }
 
-        for (let fileItem of this._fileItems) {
+        for (let [fileUri, fileItem] of this._fileItems) {
             if (fileItem.savedCoordinates !== null)
                 continue;
             if (fileItem.state != FileItem.State.NORMAL)
                 continue;
+            if (!showHidden && fileItem.isHidden)
+                continue;
             this._addFileItemCloseTo(fileItem);
         }
 
@@ -505,6 +626,15 @@ var DesktopManager = class {
         return GLib.SOURCE_REMOVE;
     }
 
+    doRename() {
+        if (this._selection.size != 1)
+            return;
+
+        let item = [...this._selection][0];
+        if (item.canRename())
+            item.doRename();
+    }
+
     doOpen() {
         for (let fileItem of this._selection)
             fileItem.doOpen();
@@ -527,6 +657,7 @@ var DesktopManager = class {
     }
 
     _onFileItemSelected(fileItem, keepCurrentSelection, addToSelection) {
+
         if (!keepCurrentSelection && !this._inDrag)
             this.clearSelection();
 
@@ -535,14 +666,13 @@ var DesktopManager = class {
         else
             this._selection.delete(fileItem);
 
-        this._fileItems.forEach(f => f.isSelected = this._selection.has(f));
+        for (let [fileUri, fileItem] of this._fileItems)
+            fileItem.isSelected = this._selection.has(fileItem);
     }
 
     clearSelection() {
-        for (let fileItem of this._fileItems) {
+        for (let [fileUri, fileItem] of this._fileItems)
             fileItem.isSelected = false;
-        }
-
         this._selection = new Set();
     }
 
@@ -568,10 +698,21 @@ var DesktopManager = class {
             this._monitorDesktopDir.cancel();
         this._monitorDesktopDir = null;
 
+        if (this.settingsId)
+            Prefs.settings.disconnect(this.settingsId);
+        this.settingsId = 0;
+        if (this.gtkSettingsId)
+            Prefs.gtkSettings.disconnect(this.gtkSettingsId);
+        this.gtkSettingsId = 0;
+
         if (this._layoutChildrenId)
             GLib.source_remove(this._layoutChildrenId);
         this._layoutChildrenId = 0;
 
+        if (this._deleteChildrenId)
+            GLib.source_remove(this._deleteChildrenId);
+        this._deleteChildrenId = 0;
+
         if (this._monitorsChangedId)
             Main.layoutManager.disconnect(this._monitorsChangedId);
         this._monitorsChangedId = 0;
@@ -585,11 +726,13 @@ var DesktopManager = class {
 
         this._rubberBand.destroy();
 
+        if (this._queryFileInfoCancellable)
+            this._queryFileInfoCancellable.cancel();
+
         Object.values(this._desktopGrids).forEach(grid => grid.actor.destroy());
         this._desktopGrids = {}
     }
-};
-Signals.addSignalMethods(DesktopManager.prototype);
+});
 
 function forEachBackgroundManager(func) {
     Main.layoutManager._bgManagers.forEach(func);
